#pragma once
#include <vector>
#include <map>
#include <tuple>
#include <algorithm>
#include <stack>

constexpr int NUM_CHARS = 256;

namespace Ukkonen
{
	template<int N>
	class SuffixTreeNode
	{
		public:
			SuffixTreeNode()
			{
				static_assert(N > 2, "The alphabet must contain at least 2 different characters!");
			}

			SuffixTreeNode(int l, int r)
			{
				static_assert(N > 2, "The alphabet must contain at least 2 different characters!");
				start = l;
				end = r;
			}

			SuffixTreeNode(SuffixTreeNode&&) noexcept = default;
			SuffixTreeNode& operator=(SuffixTreeNode&&) = default;
			SuffixTreeNode& operator=(const SuffixTreeNode&) = default;

			int start = -1;
			int end = -1;
			SuffixTreeNode<N>* suffixLink = NULL;
			std::map<int, SuffixTreeNode<N>*> branches;

	};

	template<int N>
	class SuffixTree
	{
		public:
			SuffixTree(int str[], int len, bool init = true)
			{
				static_assert(N > 2, "The alphabet must contain at least 2 different characters!");
				stringContainer.reserve(len + 1);
				stringContainer.insert(stringContainer.end(), str, str + len);
				stringContainer.push_back(N);
				stringLength = (int)stringContainer.size();
				if (init)
				{
					initEmptyTree();
					buildTree();
					finalizeTree();
				}
			}

			SuffixTree(const std::vector<int>& str, bool init = true)
			{
				static_assert(N > 2, "The alphabet must contain at least 2 different characters!");
				stringContainer.reserve(str.size() + 1);
				stringContainer.insert(stringContainer.end(), str.begin(), str.end());
				stringContainer.push_back(N);
				stringLength = (int)stringContainer.size();
				if (init)
				{
					initEmptyTree();
					buildTree();
					finalizeTree();
				}
			}

		protected:
			SuffixTreeNode<N>* root = NULL;
			SuffixTreeNode<N>* auxiliaryState = NULL;
			std::vector<int> stringContainer;
			int stringLength = -1;

			std::vector<SuffixTreeNode<N>> nodes;

			void initEmptyTree()
			{
				nodes.reserve(2 * stringLength);
				nodes.emplace_back();
				nodes.emplace_back();

				root = &nodes[0];
				auxiliaryState = &nodes[1];
				root->suffixLink = auxiliaryState;

				setupAuxiliary();
			}

			virtual void setupAuxiliary()
			{
				for (int i = 0; i < N + 1; ++i)
					auxiliaryState->branches[i] = root;
			}

			void buildTree()
			{
				if (stringLength == -1) return;

				SuffixTreeNode<N>* activeNode = root;
				int k = 0;
				for (int i = 0; i < stringLength; ++i)
				{
					k = updateTree(activeNode, k, i);
					std::tie(activeNode, k) = canonize(activeNode, k, i);
					//displayTree(i);
				}
				//displayTree(stringLength - 1);
			}

			int updateTree(SuffixTreeNode<N>*& currentNode, int k, int i)
			{
				SuffixTreeNode<N>* oldr = root;
				SuffixTreeNode<N>* r = NULL;
				bool endPoint = false;
				std::tie(endPoint, r) = testAndSplit(currentNode, k, i - 1, stringContainer[i]);

				while (!endPoint)
				{
					//nodes.push_back(SuffixTreeNode<N>(i, stringLength));
					nodes.emplace_back(i, stringLength);
					r->branches[stringContainer[i]] = &nodes[nodes.size() - 1];

					if (oldr != root)
						oldr->suffixLink = r;

					oldr = r;
					std::tie(currentNode, k) = canonize(currentNode->suffixLink, k, i - 1);
					std::tie(endPoint, r) = testAndSplit(currentNode, k, i - 1, stringContainer[i]);
				}

				if (oldr != root)
					oldr->suffixLink = currentNode;

				return k;
			}

			std::pair<bool, SuffixTreeNode<N>*> testAndSplit(SuffixTreeNode<N>* currentNode, int k, int p, int symbol)
			{
				if (k <= p)
				{
					SuffixTreeNode<N>* transition = currentNode->branches[stringContainer[k]];

					if (symbol == stringContainer[transition->start + p - k + 1])
						return { true, currentNode };
					else
					{
						//nodes.push_back(SuffixTreeNode<N>(transition->start, transition->start + p - k));
						nodes.emplace_back(transition->start, transition->start + p - k);
						SuffixTreeNode<N>* r = &nodes[nodes.size() - 1];
						currentNode->branches[stringContainer[k]] = r;
						r->branches[stringContainer[transition->start + p - k + 1]] = transition;
						transition->start = r->end + 1;
						return { false, r };
					}
				}
				else
					return { currentNode->branches.find(symbol) != currentNode->branches.end(), currentNode };
			}

			std::pair<SuffixTreeNode<N>*, int> canonize(SuffixTreeNode<N>* currentNode, int k, int p)
			{
				if (p < k) return { currentNode, k };

				SuffixTreeNode<N>* transition = currentNode->branches[stringContainer[k]];
				int temp = transition->end;

				while (temp - transition->start <= p - k)
				{
					k = k + temp - transition->start + 1;
					currentNode = transition;
					if (k <= p)
					{
						transition = currentNode->branches[stringContainer[k]];
						temp = transition->end;
						if (temp == -1 && transition->start != -1) temp = p;
					}
				}

				return { currentNode, k };
			}

			void finalizeTree()
			{
				for (int i = 2; i < (int)nodes.size(); ++i)
				{
					finalizeNode(i);
				}
			}

			virtual void finalizeNode(int i)
			{
				if (nodes[i].end == stringLength) nodes[i].end = stringLength - 1;
			}

			void recursivelyFillArray(SuffixTreeNode<N>* node, int length, int& index, std::vector<int>& sa) const
			{
				length += node->end - node->start + 1;
				if (node->branches.size() == 0)
					sa[index++] = stringLength - length;
				else
				{
					if (node->branches.find(N) != node->branches.end())
						recursivelyFillArray(node->branches[N], length, index, sa);
					for (std::pair<int, SuffixTreeNode<N>*> i : node->branches)
						if (i.first != N)
							recursivelyFillArray(i.second, length, index, sa);
				}
			}

		public:
			std::vector<int> makeSuffixArray() const
			{
				std::vector<int> ret(stringLength);
				int len = 0;
				int index = 0;
				//if (root->branches.find(N) != root->branches.end())
				//	recursivelyFillArray(root->branches[N], len, index, ret);
				//for (std::pair<int, SuffixTreeNode<N>*> i : root->branches)
				//	if (i.first != N)
				//		recursivelyFillArray(i.second, len, index, ret);
				struct MyPair
				{
					MyPair(int l, SuffixTreeNode<N>* n) : length(l), node(n) { }
					int length = 0;
					SuffixTreeNode<N>* node = 0;
				};
				std::stack<MyPair> temp;
				std::stack<MyPair> mainStack;

				mainStack.push(MyPair(0, root));

				while (!mainStack.empty())
				{
					MyPair p = mainStack.top();
					mainStack.pop();
					int length = p.length + p.node->end - p.node->start + 1;

					if ((int)p.node->branches.size() == 0)
						ret[index++] = stringLength - length + 1;
					else
					{
						if (p.node->branches.find(N) != p.node->branches.end())
							temp.push(MyPair(length, p.node->branches[N]));
						for (std::pair<int, SuffixTreeNode<N>*> i : p.node->branches)
							if (i.first != N)
								temp.push(MyPair(length, i.second));
						
						while (!temp.empty())
						{
							mainStack.push(temp.top());
							temp.pop();
						}
					}
				}

				return ret;
			}

			std::vector<int> makeSuffixArray2() const
			{
				std::vector<int> ret(stringLength);
				int len = 0;
				int index = 0;
				if (root->branches.find(N) != root->branches.end())
					recursivelyFillArray(root->branches[N], len, index, ret);
				for (std::pair<int, SuffixTreeNode<N>*> i : root->branches)
					if (i.first != N)
						recursivelyFillArray(i.second, len, index, ret);
				
				return ret;
			}
	};
}

//
//funkcija, ki izrauna BWT
//numlst - niz, ki ga transformiramo
//index - indeks BWT
//
std::vector<int> BwtEncode(const std::vector<int>& data, int& index)
{
	//konstruiramo priponsko polje
	Ukkonen::SuffixTree<NUM_CHARS> st(data);
	std::vector<int> sa = st.makeSuffixArray();
	//std::vector<int> sa2 = st.makeSuffixArray2();
	//std::ifstream vhod("sa.txt");
	//int c;
	//int e = 0;
	//for (int i = 0; i < (int)sa.size(); ++i)
	//{
	//	vhod >> c;
	//	if (c != sa[i]) ++e;
	//}
	//shranimo znak, ki bo zamenjal $
	int first = data[sa[0] - 1];
	index = -1;

	std::vector<int> bwt(sa.size() - 1);

	//izraunamo BWT
	for (int i = 1; i < (int)sa.size(); ++i)
	{
		if (sa[i] == 0)
		{
			bwt[i - 1] = first;
			index = i - 1;
		}
		else
		{
			bwt[i - 1] = data[sa[i] - 1];
		}
	}

	return bwt;
}

//
//funkcija, ki dekodira BWT
//bwt - rezultat BWT nad vhodnim nizom
//index - indeks BWT
//
std::vector<int> BwtDecode(const std::vector<int>& bwt, int index)
{
	//$ vstavimo nazaj na poloaj, doloen z indeksom
	std::vector<int> temp(bwt.size() + 1, 0);
	temp[0] = bwt[index] + 1;
	std::vector<int> ibwt(bwt.size(), 0);
	int numIndices = NUM_CHARS + 1;
	//pretejemo, koliko razlinih znakov nastopa v nizu
	for (int i = 0; i < (int)bwt.size(); ++i)
		temp[i + 1] = bwt[i] + 1;

	temp[index + 1] = 0;

	//polje C pove, koliko enakih znakov se je pred tem indeksom e pojavilo
	//polje K pove, kolikokrat se kak znak pojavi
	std::vector<int> C(temp.size(), 0);
	std::vector<int> K(numIndices, 0);
	for (int i = 0; i < (int)temp.size(); ++i)
		C[i] = K[temp[i]]++;

	//M hrani indeks prve pojavitve vsakega znaka
	std::vector<int> M(numIndices, 0);
	M[0] = 0;
	for (int i = 1; i < numIndices; ++i)
		M[i] = M[i - 1] + K[i - 1];

	++index;
	index = C[index] + M[temp[index]];
	//s pomonimi polji izvedemo inverzno BWT
	for (int i = (int)bwt.size() - 1; i > -1; --i)
	{
		ibwt[i] = temp[index] - 1;
		index = C[index] + M[temp[index]];
	}

	return ibwt;
}